//
// Copyright 2007, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.

#ifndef myOstream_header_included
#define myOstream_header_included

//
//.  This file defines an ostream-like class from scratch.  It is not a piece of production
//.  code so much as it is an example of how you might implement an ostream class in ROM.
//.  Although, this particular example doesn't actually accomplish that.  
//.
//.  More to the point, a myOstream object ACTS like a std::ostream object, but is not in
//.  anyway related to it.  This code is effectively a demonstration of how to get ostream
//.  functionality without the baggage of std::ostream.
//.
//.  Note that if you are not in fact trying to produce a super small program with no
//.  excess code inclusion at link time, you might consider a different approach to creating
//.  your own stream class.  The std stream functionality lets you derive your own
//.  stream buffer class.  In many production s/w systems, that is the approach you should be
//.  using for maintenance reasons.
//.
//.  This header file also defines myOcout to aid in testing and message generation.  myOcout
//.  is a myOstream which sends its data to cout.
//

// define the following symbol yourself, before including this header file,
// if you want this ostream variant to work with std::ostream's manipulators -- like std::endl,
// std::setw, std::setfill, std::hex, etc..  Otherwise, you get no support for it.

#ifndef MYOSTREAM_PLAYS_NICE_WITH_STD_OSTREAM  
#define MYOSTREAM_PLAYS_NICE_WITH_STD_OSTREAM   0
#endif

#include <string>

#if MYOSTREAM_PLAYS_NICE_WITH_STD_OSTREAM
#include <iosfwd>
#endif

class myOstream;

extern myOstream &myOendl(myOstream&);  // same as std::endl but only for this kind of stream
extern myOstream &myOflush(myOstream&);
extern myOstream &myOhex(myOstream&);
extern myOstream &myOdec(myOstream&);
extern myOstream &myOoct(myOstream&);
extern myOstream &myOshowbase(myOstream&);
extern myOstream &myOnoshowbase(myOstream&);

struct myOstream
  // myOstream is meant to be the interface definition for a family of derived classes which
  // implement std::ostream like functionality.  The base class, myOstream, implements all the
  // io formatting and buffering, but not the actual transfer of data to a file handle.  The
  // derived class is meant to handle the part of the task.
  //
  // User's should program to the myOstream interface, but understand that when they create a
  // stream, they will be creating an object of a derived class.
  //
  // To implement a stream of your own, derive from this class and add appropriate constructors
  // and data members.  Make your that you implement a destructor and that your destructor
  // calls flush()!  This can't be done automatically because in the base class destructor
  // because myOstream::flush() calls your virtual method, WriteFile(), which the compiler will
  // have taken out of the virtual table by the time ~myOstream() runs.
  //
  // You must implement WriteFile(), in what ever manner you see fit.  The parameters to the
  // constructor for your derived class must also handle any initialization parameters -- like
  // the name of a file to write too, open modes, etc.  If you are sending the output to 
  // RS232 port, you must include port ids, speeds, etc in the constructor parms -- at least
  // that's what I'd do.
{
public:

  virtual bool WriteFile(char const *buffer, size_t length, size_t *actualBytesRead) = 0;
    // Performs the lowest level of file write operation.  The return value is false if
    // the number of bytes written does not equal the requested length.  The number actually
    // written is returned via the pointer, actualBytesRead.

  void flush();
    // Ensures that any pending data gets written using the WriteFile() method.

  size_t  tellp(); // gets number of chars previously output.


  // Now define the output operators for this kind of stream.  The built in types all need there
  // own member function to handle their io and all other types need to create a operator
  // function themselves.

   myOstream &operator<< (unsigned char c);     
   myOstream &operator<< (unsigned short s);    
   myOstream &operator<< (unsigned int i);      
   myOstream &operator<< (unsigned long l);     
   myOstream &operator<< (int i);      
   myOstream &operator<< (int long);      
   myOstream &operator<< (char c);      
   myOstream &operator<< (char const *);        
   myOstream &operator<< (std::string const &); 
   myOstream &operator<< (float const &);       
   myOstream &operator<< (double const &);       
   myOstream &put        (char c);

  template<class T> myOstream & operator<< (T const &); // handle all remaining types
    //
    // All objects for which no specific operator<< is provided as a special case, above, 
    // are handled by this template signature.  If you get a link error telling you that 
    // a given signature is not found in your object modules, then add one yourself using 
    // template specialization like this:
    //
    //    template<> myOstream& myOstream::operator<< <YourType> ( YourType const & obj )
    //    { ... ; return *this; }
    //
    // Note that you can still define global free function operator<<'s like this:
    //
    //    myOstream& operator<< (myOstream &s, YourType const &obj)
    //    { ... ; return s; }
    //
    // That approach just pollutes the name space more.

   myOstream &operator<< ( myOstream & (*ptr)(myOstream &) ) { return (*ptr)(*this); }
     // handle stream manipulations of various kinds:  myOendl, myOflush, enable std::endl & flush


#if MYOSTREAM_PLAYS_NICE_WITH_STD_OSTREAM

   myOstream &operator<< ( std::ostream & (*ptr)(std::ostream& ) ); 
     // for std::endl and std::flush

   myOstream &operator<< ( std::ios_base &(*ptr)(std::ios_base&) ); 
     // for std::hex, dec, etc
#endif

   virtual ~myOstream()
   {
     // don't put a flush() call here -- let the derived classes do it because the flush()
     // function calls the WriteFile() function and the vtable in the class is empty by the
     // time you get here.
   }

   // the following variables are used in the formatting output -- and the
   // manipulators and stream functions modify them

   std::string outputBuffer_;

   enum outputBaseType { hexOutput, decimalOutput, octalOutput };

   outputBaseType outputBase; // defaults to decimal

   char padCharacter;  // defaults to ' '

   size_t minWidth; // defaults to 0 -- any output will be at least this while

   size_t num_put_; // number of chars put using WriteFile

   bool showBase_;

   myOstream()
   : outputBase(decimalOutput),
     padCharacter(' '),
     minWidth(0),
     num_put_(0),
     showBase_(0)
   {
   }


   struct fillcharT { char fc; fillcharT(char c): fc(c) {} };

   myOstream& operator<<(fillcharT fc) { padCharacter = fc.fc; return *this; }
     //.  This function is invoked when you use the following syntax:
     //.    os << myOsetfill('x') << ...


   struct widthT { size_t fw; widthT(size_t w): fw(w) {} };

   myOstream& operator<<(widthT fw) { minWidth = fw.fw; return *this; }
     //.  This function is invoked when you use the following syntax:
     //.    os << myOsetw(12) << ...



private:
  void dumpWithPad(char const *s, size_t stringLength); // append s to output buffer, add needed padding, set minWidth to 0.

};

inline myOstream::fillcharT  myOsetfill(char c)    { return c; } 
     //.  global free function that sets stream padding character

inline myOstream::widthT     myOsetw(size_t w)     { return w; } 
     // global free function ath sets stream field width


struct myCoutStream
: public myOstream
{
  // A myCoutStream implements a myOstream that send all its data to cout.
  // Only one of these things needs to be created for a program, but creating many 
  // won't hurt anything.

  virtual bool WriteFile(char const *buffer, size_t length, size_t *actualBytesRead);

  ~myCoutStream()
  {
    flush(); // force pending output
  }

};

extern myCoutStream myOcout; // the object has 0 size and could be statically declared in the
                             // the header file if you were lazy -- but every module including
                             // myOcout would get stuck with it...

struct myOfileStream
: public myOstream
{
  // a myOfileStream implements a myOstream that send all its data to a file whose name
  // is specified at the time of construction.

  struct impl;

  impl *implementation_;

  myOfileStream(char const *name);  // opens with O_CREAT|O_TRUNC|O_RDWR, with permissions
                                    // for all to read and write.
 ~myOfileStream();

  virtual bool WriteFile(char const *buffer, size_t length, size_t *actualBytesRead);

};



#endif
